
import math
import numpy as np
from typing import Union


__all__ = [
    'psi',
    'chi',
    'psichi',
]


def offset_iterator(base_frequency, N, maximum_offset=None, maximum_frequency=None, non_negative: bool = False):
    if N < 0:
        # assert maximum_offset == 0
        if maximum_frequency is not None and math.fabs(base_frequency) <= maximum_frequency:
            yield 0
    else:
        assert maximum_frequency is not None or maximum_offset is not None
        
        if non_negative:
            minimum_frequency = 0
        else:
            minimum_frequency = -maximum_frequency if maximum_frequency is not None else None

        def round(x):
            if x > 0:
                return int(math.floor(x))
            else:
                return int(math.ceil(x))
            
        if maximum_frequency is not None:
            min_offset = (minimum_frequency - base_frequency) / N
            max_offset = (maximum_frequency - base_frequency) / N
        else:
            min_offset = -10000
            max_offset = 10000
            
        if maximum_offset is not None:
            min_offset = max(min_offset, -maximum_offset)
            max_offset = min(max_offset, maximum_offset)
        
        min_offset = math.ceil(min_offset)
        max_offset = math.floor(max_offset)
        
        for j in range(min_offset, max_offset+1):
            yield j


def psi(theta: Union[np.ndarray, float],
        k: int = 1,
        gamma: float = 0.,
        out: np.ndarray = None) -> np.ndarray:
    
    # rotation matrix of frequency k corresponding to the angle theta

    if isinstance(theta, float):
        theta = np.array(theta)

    k = np.array(k, copy=False).reshape(-1, 1)
    gamma = np.array(gamma, copy=False).reshape(-1, 1)
    theta = theta.reshape(1, -1)

    x = k * theta + gamma

    cos, sin = np.cos(x), np.sin(x)

    if out is None:
        out = np.empty((2, 2, x.shape[0], x.shape[-1]))

    out[0, 0, ...] = cos
    out[0, 1, ...] = -sin
    out[1, 0, ...] = sin
    out[1, 1, ...] = cos
    
    return out


def chi(s: int, out: np.ndarray = None) -> np.ndarray:
    # the orthonormal matrix associated to the flip
    
    s = -1 * (s > 0) + (s <= 0)

    s = np.array(s, copy=False).reshape(-1, 1)

    if out is None:
        out = np.empty((2, 2, s.shape[0], 1))
    
    out[0, 0, ...] = 1
    out[0, 1, ...] = 0
    out[1, 0, ...] = 0
    out[1, 1, ...] = s
    return out


def psichi(theta: Union[np.ndarray, float], s: int, k: int = 1, gamma: float = 0., out: np.ndarray = None) -> np.ndarray:
    # equal to the matrix multiplication of psi(theta, k, gamma) and chi(s) along the first 2 axis
    
    if isinstance(theta, float):
        theta = np.array(theta)
    
    s = -1 * (s > 0) + (s <= 0)
    
    s = np.array(s, copy=False).reshape(-1, 1)
    k = np.array(k, copy=False).reshape(-1, 1)
    gamma = np.array(gamma, copy=False).reshape(-1, 1)
    theta = theta.reshape(1, -1)
    
    x = k * theta + gamma
    
    cos, sin = np.cos(x), np.sin(x)
    
    if out is None:
        out = np.empty((2, 2, x.shape[0], x.shape[-1]))
    
    out[0, 0, ...] = cos
    out[0, 1, ...] = -s*sin
    out[1, 0, ...] = sin
    out[1, 1, ...] = s*cos
    
    return out


if __name__ == "__main__":

    def test(N, b, MF):
        fs = []
        for f in range(-MF, MF+1):
            if (f-b) % N == 0:
                fs.append(f)
            
        fs2 = [b + j*N for j in offset_iterator(b, N, None, MF)]
        
        assert sorted(fs) == sorted(fs2), (N, b, MF, fs, fs2)


    for N in range(1, 30):
        for b in range(N):
            for MF in range(1, 40):
                test(N, b, MF)
    
    #points = np.array([[0., 0.], [0., 1.], [1., 1.], [-1, 0.]]).T
    points = np.random.randn(2, 50)
    angles = np.arctan2(points[1, :], points[0, :])
    gamma = np.array([0., 0., np.pi / 2])
    k = np.array([1, 2, 4])
    s = np.array([1, 0, 0])
    
    p = psi(angles, k, gamma).transpose(2, 3, 0, 1)
    c = chi(s).transpose(2, 3, 0, 1)
    pc = psichi(angles, s, k, gamma).transpose(2, 3, 0, 1)
    
    assert np.allclose(p@c, pc)





from .basis import EmptyBasisException, KernelBasis, AdjointBasis, UnionBasis

from .spaces import *

from .steerable_basis import SteerableKernelBasis

from .polar_basis import GaussianRadialProfile, SphericalShellsBasis

from .wignereckart_solver import WignerEckartBasis, RestrictedWignerEckartBasis
from .sparse_basis import SparseSteerableBasis

from .r2 import *
from .r3 import *


import escnn.group


def kernels_on_point(in_repr: escnn.group.Representation, out_repr: escnn.group.Representation) -> KernelBasis:
    r"""

    Args:
        in_repr (Representation): the representation specifying the transformation of the input feature field
        out_repr (Representation): the representation specifying the transformation of the output feature field
        
    """
    assert in_repr.group == out_repr.group
    
    group = in_repr.group
    
    basis = SteerableKernelBasis(
        PointRn(1, group), in_repr, out_repr,
        WignerEckartBasis,
        harmonics=[group.trivial_representation.id]
    )
    
    return basis


__all__ = [
    "EmptyBasisException",
    "KernelBasis",
    # General Bases
    'SteerableKernelBasis',
    "WignerEckartBasis",
    "RestrictedWignerEckartBasis",
    "SparseSteerableBasis",
    "AdjointBasis",
    "UnionBasis",
    # Specific Bases
    "SphericalShellsBasis",
    'GaussianRadialProfile',
    # Space Isomorphisms
    'SpaceIsomorphism',
    'CircleO2',
    'CircleSO2',
    'SphereO3',
    'SphereSO3',
    'PointRn',
    'Icosidodecahedron',
    'Dodecahedron',
    'Icosahedron',
    # Generic group acting on R^0
    "kernels_on_point",
    # R2 bases
    "kernels_Flip_act_R2",
    "kernels_DN_act_R2",
    "kernels_O2_act_R2",
    "kernels_Trivial_act_R2",
    "kernels_CN_act_R2",
    "kernels_SO2_act_R2",
    "kernels_SO2_subgroup_act_R2",
    "kernels_O2_subgroup_act_R2",
    # R3 bases
    "kernels_O3_act_R3",
    "kernels_SO3_act_R3",
    "kernels_SO3_subgroup_act_R3",
    "kernels_O3_subgroup_act_R3",
    "kernels_Ico_act_R3",
    "kernels_Octa_act_R3",
    "kernels_Tetra_act_R3",
    "kernels_FullIco_act_R3",
    "kernels_FullOcta_act_R3",
    "kernels_FullTetra_act_R3",
    "kernels_Pyrito_act_R3",
    "kernels_SO2_act_R3",
    "kernels_CN_act_R3",
    "kernels_O2_conical_act_R3",
    "kernels_DN_conical_act_R3",
    "kernels_O2_dihedral_act_R3",
    "kernels_DN_dihedral_act_R3",
    "kernels_Inv_act_R3",
    "kernels_Trivial_act_R3",
    'kernels_aliased_Ico_act_R3_dodecahedron',
    'kernels_aliased_Ico_act_R3_icosidodecahedron',
    'kernels_aliased_Ico_act_R3_icosahedron',
]





from escnn.nn import FieldType
from escnn.group import Representation
from escnn.kernels import KernelBasis

from .rd_point_convolution import _RdPointConv
from .r2_point_convolution import compute_basis_params

from typing import Callable, Tuple, Dict, Union, List

__all__ = ["R3PointConv"]


class R3PointConv(_RdPointConv):
    
    def __init__(self,
                 in_type: FieldType,
                 out_type: FieldType,
                 groups: int = 1,
                 bias: bool = True,
                 sigma: Union[List[float], float] = None,
                 width: float = None,
                 n_rings: int = None,
                 frequencies_cutoff: Union[float, Callable[[float], int]] = None,
                 rings: List[float] = None,
                 basis_filter: Callable[[dict], bool] = None,
                 recompute: bool = False,
                 ):

        basis_filter, self._rings, self._sigma, self._maximum_frequency = compute_basis_params(
            frequencies_cutoff, rings, sigma, width, n_rings, basis_filter
        )

        super(R3PointConv, self).__init__(
            in_type, out_type,
            d=3,
            groups=groups,
            bias=bias,
            basis_filter=basis_filter,
            recompute=recompute
        )

    def _build_kernel_basis(self, in_repr: Representation, out_repr: Representation) -> KernelBasis:
        return self.space.build_kernel_basis(in_repr, out_repr,
                                             self._sigma, self._rings,
                                             maximum_frequency=self._maximum_frequency
                                             )






import numpy as np

from .basis import KernelBasis, EmptyBasisException
from .spaces import SpaceIsomorphism

from escnn.group import Group
from escnn.group import IrreducibleRepresentation
from escnn.group import Representation

from typing import Type, Union, Tuple, Dict, List, Iterable, Callable, Set
from abc import ABC, abstractmethod
from collections import defaultdict


class IrrepBasis(KernelBasis):
    
    def __init__(self,
                 X: SpaceIsomorphism,
                 in_irrep: Union[IrreducibleRepresentation, Tuple],
                 out_irrep: Union[IrreducibleRepresentation, Tuple],
                 js: List[Tuple],
                 dim: int):
        r"""

        Abstract class for bases implementing the kernel constraint solutions associated to irreducible input and output
        representations.

        Args:
            in_irrep:
            out_irrep:
            dim:
        """
        self.X = X
        assert in_irrep.group == out_irrep.group
        self.group = in_irrep.group
        self.in_irrep = self.group.irrep(*self.group.get_irrep_id(in_irrep))
        self.out_irrep = self.group.irrep(*self.group.get_irrep_id(out_irrep))
        
        self.js = js

        super(IrrepBasis, self).__init__(dim, (out_irrep.size, in_irrep.size))

        self._start_index = {}
        idx = 0
        for _j in self.js:
            self._start_index[_j] = idx
            idx += self.dim_harmonic(_j)

    def sample(self, points: np.ndarray, out: np.ndarray = None) -> np.ndarray:
        r"""

        Sample the continuous basis elements on the discrete set of points in ``points``.
        Optionally, store the resulting multidimentional array in ``out``.

        ``points`` must be an array of shape `(d, N)`, where `N` is the number of points.

        Args:
            points (~numpy.ndarray): points where to evaluate the basis elements
            out (~numpy.ndarray, optional): pre-existing array to use to store the output

        Returns:
            the sampled basis

        """
    
        assert len(points.shape) == 2
        S = points.shape[1]
    
        if out is None:
            out = np.empty((self.shape[0], self.shape[1], self.dim, S))
    
        assert out.shape == (self.shape[0], self.shape[1], self.dim, S)

        B = 0
        harmonics = {}
        outs = {}
        for b, j in enumerate(self.js):
    
            Ys = self.X.basis(points, j)
    
            harmonics[j] = Ys
            
            outs[j] = out[:, :, B:B + self.dim_harmonic(j), :]
            B += self.dim_harmonic(j)

        self.sample_harmonics(harmonics, outs)
        return out
    
    @abstractmethod
    def sample_harmonics(self, points: Dict[Tuple, np.ndarray], out: Dict[Tuple, np.ndarray] = None) -> Dict[Tuple, np.ndarray]:
        raise NotImplementedError()

    @abstractmethod
    def dim_harmonic(self, j: Tuple) -> int:
        raise NotImplementedError()
    
    @abstractmethod
    def attrs_j_iter(self, j: Tuple) -> Iterable:
        raise NotImplementedError()
    
    @abstractmethod
    def attrs_j(self, j: Tuple, idx) -> Dict:
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def _generator(cls,
                   X: SpaceIsomorphism,
                   psi_in: Union[IrreducibleRepresentation, Tuple],
                   psi_out: Union[IrreducibleRepresentation, Tuple],
                   harmonics: List[Tuple] = None,
                   **kwargs
    ) -> 'IrrepBasis':
        raise NotImplementedError()


class SteerableKernelBasis(KernelBasis):
    
    def __init__(self,
                 X: SpaceIsomorphism,
                 in_repr: Representation,
                 out_repr: Representation,
                 irreps_basis: Type[IrrepBasis],
                 harmonics: Union[List, Set] = None, # optionally filter only some harmonics
                 **kwargs):
        r"""
        
        Implements a general basis for the vector space of equivariant kernels over the homogeneous space :math:`X`.
        A :math:`G`-equivariant kernel :math:`\kappa`, mapping between an input field, transforming under
        :math:`\rho_\text{in}` (``in_repr``), and an output field, transforming under  :math:`\rho_\text{out}`
        (``out_repr``), satisfies the following constraint:
        
        .. math ::
            
            \kappa(gx) = \rho_\text{out}(g) \kappa(x) \rho_\text{in}(g)^{-1} \qquad \forall g \in G, \forall x \in X
        
        As the kernel constraint is a linear constraint, the space of equivariant kernels is a vector subspace of the
        space of all convolutional kernels. It follows that any equivariant kernel can be expressed in terms of a basis
        of this space.
        
        This class solves the kernel constraint for two arbitrary representations by combining the solutions of the
        kernel constraints associated to their :class:`~escnn.group.IrreducibleRepresentation` s.
        In order to do so, it relies on ``irreps_basis`` which solves individual irreps constraints. ``irreps_basis``
        must be a class (subclass of :class:`~escnn.kernels.IrrepsBasis`) which builds a basis for equivariant
        kernels associated with irreducible representations when instantiated.
        
        The groups :math:`G` which are currently implemented are origin-preserving isometries (what are called
        structure groups, or sometimes gauge groups, in the language of
        `Gauge Equivariant CNNs <https://arxiv.org/abs/1902.04615>`_ ).
        The origin-preserving isometries of :math:`\R^d` are subgroups of :math:`O(d)`, i.e. reflections and rotations.
        Therefore, equivariance does not enforce any constraint on the radial component of the kernels.
        Hence, this class only implements a basis for the angular part of the kernels.
        
        In order to build a complete basis of kernels, you should combine this basis with a basis which defines the
        radial profile (such as :class:`~escnn.kernels.GaussianRadialProfile`) through
        :class:`~escnn.kernels.SphericalShellsBasis`.
        
        .. math::
            
            \mathcal{B} = \left\{ b_i (r) :=  \exp \left( \frac{ \left( r - r_i \right)^2}{2 \sigma_i^2} \right) \right\}_i
        
        .. warning ::
            
            Typically, the user does not need to manually instantiate this class.
            Instead, we suggest to use the interface provided in :doc:`escnn.gspaces`.
        
        Args:
            X (SpaceIsomorphism): the base space where the steerable kernel is defined
            in_repr (Representation): Representation associated with the input feature field
            out_repr (Representation): Representation associated with the output feature field
            irreps_basis (class): class defining the irreps basis. This class is instantiated for each pair of irreps to solve all irreps constraints.
            harmonics (optional): selects only a subset of the harmonics to use.
            **kwargs: additional arguments used when instantiating ``irreps_basis``
            
        """
        
        assert in_repr.group == out_repr.group
        
        self.X = X
        self.in_repr = in_repr
        self.out_repr = out_repr
        group = in_repr.group
        self.group = group
        
        self._irrep_basis = irreps_basis
        self._irrep__basis_kwargs = kwargs

        A_inv = np.array(in_repr.change_of_basis_inv, copy=True)
        B = np.array(out_repr.change_of_basis, copy=True)
        
        # A_inv = in_repr.change_of_basis_inv
        # B = out_repr.change_of_basis

        if not np.allclose(A_inv, np.eye(in_repr.size)):
            self.A_inv = A_inv
        else:
            self.A_inv = None
            
        if not np.allclose(B, np.eye(out_repr.size)):
            self.B = B
        else:
            self.B = None

        # Dict[Tuple, IrrepsBasis]:
        self.irreps_bases = {}
        
        js = set()
        
        # loop over all input irreps
        for i_irrep_id in set(in_repr.irreps):
            # loop over all output irreps
            for o_irrep_id in set(out_repr.irreps):
        
                try:
                    # retrieve the irrep intertwiner basis
                    basis = irreps_basis._generator(self.X, i_irrep_id, o_irrep_id, harmonics=harmonics, **kwargs)
                    assert basis.group == self.group

                    self.irreps_bases[(i_irrep_id, o_irrep_id)] = basis
                    
                    # compute the set of all harmonics needed by all irreps bases
                    # in this way, we can precompute the embedding of the points with the harmonics and reuse the same
                    # embeddings for all irreps bases
                    js.update(basis.js)

                except EmptyBasisException:
                    # if the basis is empty, skip it
                    pass
        
        if callable(harmonics):
            js = list(j for j in js if harmonics(j))
        elif isinstance(harmonics, set) or isinstance(harmonics, list):
            js = js.intersection(set(harmonics))
        
        self.js = sorted(list(js))
        if self.X.zero_harmonic in self.js:
            # make sure that the harmonic corresponding to the trivial representation (i.e. the harmonic spanning
            # the space of constant functions over the space) is the first in the list.
            self.js.remove(self.X.zero_harmonic)
            self.js = [self.X.zero_harmonic] + self.js
        
        self.bases = [[None for _ in range(len(out_repr.irreps))] for _ in range(len(in_repr.irreps))]
        
        self.in_sizes = []
        self.out_sizes = []
        # loop over all input irreps
        for ii, i_irrep_id in enumerate(in_repr.irreps):
            self.in_sizes.append(group.irrep(*i_irrep_id).size)
            
        # loop over all output irreps
        for oo, o_irrep_id in enumerate(out_repr.irreps):
            self.out_sizes.append(group.irrep(*o_irrep_id).size)

        self._dim_harmonics = defaultdict(int)
        dim = 0
        # loop over all input irreps
        for ii, i_irrep_id in enumerate(in_repr.irreps):
            # loop over all output irreps
            for oo, o_irrep_id in enumerate(out_repr.irreps):
                if (i_irrep_id, o_irrep_id) in self.irreps_bases:
                    self.bases[ii][oo] = self.irreps_bases[(i_irrep_id, o_irrep_id)]
                    dim += self.bases[ii][oo].dim
                    for j in self.bases[ii][oo].js:
                        self._dim_harmonics[j] += self.irreps_bases[(i_irrep_id, o_irrep_id)].dim_harmonic(j)
                        
        self._slices = defaultdict(dict)
        basis_count = defaultdict(int)
        in_position = 0
        for ii, in_size in enumerate(self.in_sizes):
            out_position = 0
            for oo, out_size in enumerate(self.out_sizes):
                if self.bases[ii][oo] is not None:
                    for j in self.bases[ii][oo].js:
                        self._slices[(ii, oo)][j] = (
                            out_position,
                            out_position + out_size,
                            in_position,
                            in_position + in_size,
                            basis_count[j],
                            basis_count[j] + self.bases[ii][oo].dim_harmonic(j)
                        )
                        basis_count[j] += self.bases[ii][oo].dim_harmonic(j)
                out_position += out_size
            in_position += in_size

        super(SteerableKernelBasis, self).__init__(dim, (out_repr.size, in_repr.size))
        
    def dim_harmonic(self, j: Tuple) -> int:
        return self._dim_harmonics[j]

    def compute_harmonics(self, points: np.ndarray) -> Dict[Tuple, np.ndarray]:
        harmonics = {}
        for j in self.js:
            Ys = self.X.basis(points, j)
            harmonics[j] = Ys

        return harmonics

    def sample(self, points: np.ndarray, out: np.ndarray = None) -> np.ndarray:
        r"""

        Sample the continuous basis elements on the discrete set of points in ``points``.
        Optionally, store the resulting multidimentional array in ``out``.
        
        ``points`` must be an array of shape `(d, N)`, where `N` is the number of points.

        Args:
            points (~numpy.ndarray): points where to evaluate the basis elements
            out (~numpy.ndarray, optional): pre-existing array to use to store the output

        Returns:
            the sampled basis
            
        """
        assert len(points.shape) == 2

        if out is None:
            out = np.zeros((self.shape[0], self.shape[1], self.dim, points.shape[1]))
        else:
            out.fill(0.)
            
        assert out.shape == (self.shape[0], self.shape[1], self.dim, points.shape[1])

        harmonics = {}
        outs = {}
        B = 0
        for j in self.js:
            Ys = self.X.basis(points, j)
    
            harmonics[j] = Ys
    
            outs[j] = out[:, :, B:B + self.dim_harmonic(j), :]
            B += self.dim_harmonic(j)

        self.sample_harmonics(harmonics, outs)
        return out

    def sample_harmonics(self, points: Dict[Tuple, np.ndarray], out: Dict[Tuple, np.ndarray] = None) -> Dict[Tuple, np.ndarray]:
        if out is None:
            out = {
                j: np.zeros((self.shape[0], self.shape[1], self.dim_harmonic(j), points[j].shape[-1]))
                for j in self.js
            }

        for j in self.js:
            if j in out:
                assert out[j].shape == (self.shape[0], self.shape[1], self.dim_harmonic(j), points[j].shape[-1])
    
        if self.A_inv is None and self.B is None:
            out = self._sample_direct_sum(points, out=out)
        else:
            samples = self._sample_direct_sum(points)
            out = self._change_of_basis(samples, out=out)
    
        return out

    def _sample_direct_sum(self, points: Dict[Tuple, np.ndarray], out: Dict[Tuple, np.ndarray] = None) -> Dict[Tuple, np.ndarray]:
    
        if out is None:
            out = {
                j: np.zeros((self.shape[0], self.shape[1], self.dim_harmonic(j), points[j].shape[-1]))
                for j in self.js
            }
        else:
            for j in self.js:
                if j in out:
                    out[j].fill(0)
    
        for j in self.js:
            if j in out:
                assert out[j].shape == (self.shape[0], self.shape[1], self.dim_harmonic(j), points[j].shape[-1])

        for ii, in_size in enumerate(self.in_sizes):
            for oo, out_size in enumerate(self.out_sizes):
                if self.bases[ii][oo] is not None:
                    slices = self._slices[(ii, oo)]
                    
                    blocks = {
                        j: out[j][o_s:o_e, i_s:i_e, b_s:b_e, ...]
                        for j, (o_s, o_e, i_s, i_e, b_s, b_e) in slices.items()
                    }
                    self.bases[ii][oo].sample_harmonics(points, out=blocks)

        return out
    
    def _change_of_basis(self, samples: Dict[Tuple, np.ndarray], out: Dict[Tuple, np.ndarray] = None) -> Dict[Tuple, np.ndarray]:
        # multiply by the change of basis matrices to transform the irreps basis in the full representations basis

        if out is None:
            out = {j: None for j in self.js}
            
        for j in samples.keys():
            if self.A_inv is not None and self.B is not None:
                out[j] = np.einsum("no,oibp,ij->njbp", self.B, samples[j], self.A_inv, out=out[j])
            elif self.A_inv is not None:
                out[j] = np.einsum("oibp,ij->ojbp", samples[j], self.A_inv, out=out[j])
            elif self.B is not None:
                out[j] = np.einsum("no,oibp->nibp", self.B, samples[j], out=out[j])
            else:
                out[j][...] = samples[j]
        
        return out

    def __getitem__(self, idx):
        assert idx < self.dim
        
        j_idx = idx
        for j in self.js:
            dim = self.dim_harmonic(j)
            if j_idx >= dim:
                j_idx -= dim
            else:
                break
        
        assert j_idx < self.dim_harmonic(j)

        count = 0
        for ii in range(len(self.in_sizes)):
            for oo in range(len(self.out_sizes)):
                if self.bases[ii][oo] is not None:
                    dim = self.bases[ii][oo].dim_harmonic(j)

                    rel_idx = j_idx - count
                    if rel_idx >= 0 and rel_idx < dim:
                        
                        attr = dict(self.bases[ii][oo].attrs_j(j, rel_idx))
                        
                        attr["shape"] = self.bases[ii][oo].shape

                        attr["irreps_basis_idx"] = attr["idx"]
                        attr["idx"] = idx
                        attr["j"] = j
                        attr["j_idx"] = j_idx

                        attr["in_irrep"] = self.in_repr.irreps[ii]
                        attr["out_irrep"] = self.out_repr.irreps[oo]
                        
                        attr["in_irrep_idx"] = ii
                        attr["out_irrep_idx"] = oo
                        
                        return attr
                
                    count += dim

    def __iter__(self):
        idx = 0
        for j in self.js:
            j_idx = 0
            for ii in range(len(self.in_sizes)):
                for oo in range(len(self.out_sizes)):
                    basis = self.bases[ii][oo]
                    if basis is not None:
                        
                        for attr in basis.attrs_j_iter(j):
    
                            attr["shape"] = basis.shape
                            
                            attr["in_irrep"] = self.in_repr.irreps[ii]
                            attr["out_irrep"] = self.out_repr.irreps[oo]
                        
                            attr["in_irrep_idx"] = ii
                            attr["out_irrep_idx"] = oo

                            attr["j"] = j
                            attr["j_idx"] = j_idx
                            
                            attr["irreps_basis_idx"] = attr["idx"]
                            attr["idx"] = idx
                            
                            assert idx < self.dim

                            yield attr
                            
                            idx += 1
                            j_idx += 1

    def __eq__(self, other):
        if not isinstance(other, SteerableKernelBasis):
            return False
        elif self.X != other.X or self.in_repr != other.in_repr or self.out_repr != other.out_repr:
            return False
        else:
            sbk1 = sorted(self.irreps_bases.keys())
            sbk2 = sorted(other.irreps_bases.keys())
            if sbk1 != sbk2:
                return False
            
            for irreps, basis in self.irreps_bases.items():
                if basis != other.irreps_bases[irreps]:
                    return False
            
            return True

    def __hash__(self):
        h = hash(self.in_repr) + hash(self.out_repr) + hash(self.X)
        for basis in self.irreps_bases.items():
            h += hash(basis)
        return h
        
        






from abc import ABC, abstractmethod

from escnn.nn import FieldType
from escnn.nn import GeometricTensor
from escnn.group import Representation
from escnn.kernels import KernelBasis, EmptyBasisException
from escnn.gspaces import *

from ..equivariant_module import EquivariantModule

from escnn.nn.modules.basismanager import BlocksBasisSampler

from typing import Callable, Tuple, Dict, Union

import torch

import torch_geometric

from torch.nn import Parameter
import numpy as np
import math


__all__ = ["_RdPointConv"]


class _RdPointConv(torch_geometric.nn.MessagePassing, EquivariantModule, ABC):
    
    def __init__(self,
                 in_type: FieldType,
                 out_type: FieldType,
                 d: int,
                 groups: int = 1,
                 bias: bool = True,
                 basis_filter: Callable[[dict], bool] = None,
                 recompute: bool = False,
                 ):

        assert in_type.gspace == out_type.gspace
        assert isinstance(in_type.gspace, GSpace)
        assert d >= in_type.gspace.dimensionality

        super(_RdPointConv, self).__init__(aggr='mean')

        self.d = d
        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = out_type

        assert self.space.dimensionality == self.d
        
        self.groups = groups

        if groups > 1:
            # Check the input and output classes can be split in `groups` groups, all equal to each other
            # first, check that the number of fields is divisible by `groups`
            assert len(in_type) % groups == 0
            assert len(out_type) % groups == 0
            in_size = len(in_type) // groups
            out_size = len(out_type) // groups
            
            # then, check that all groups are equal to each other, i.e. have the same types in the same order
            assert all(in_type.representations[i] == in_type.representations[i % in_size] for i in range(len(in_type)))
            assert all(out_type.representations[i] == out_type.representations[i % out_size] for i in range(len(out_type)))
            
            # finally, retrieve the type associated to a single group in input.
            # this type will be used to build a smaller kernel basis and a smaller filter
            # as in PyTorch, to build a filter for grouped convolution, we build a filter which maps from one input
            # group to all output groups. Then, PyTorch's standard convolution routine interpret this filter as `groups`
            # different filters, each mapping an input group to an output group.
            in_type = in_type.index_select(list(range(in_size)))
        
        if bias:
            # bias can be applied only to trivial irreps inside the representation
            # to apply bias to a field we learn a bias for each trivial irreps it contains
            # and, then, we transform it with the change of basis matrix to be able to apply it to the whole field
            # this is equivalent to transform the field to its irreps through the inverse change of basis,
            # sum the bias only to the trivial irrep and then map it back with the change of basis
            
            # count the number of trivial irreps
            trivials = 0
            for r in self.out_type:
                for irr in r.irreps:
                    if self.out_type.fibergroup.irrep(*irr).is_trivial():
                        trivials += 1
            
            # if there is at least 1 trivial irrep
            if trivials > 0:
                
                # matrix containing the columns of the change of basis which map from the trivial irreps to the
                # field representations. This matrix allows us to map the bias defined only over the trivial irreps
                # to a bias for the whole field more efficiently
                bias_expansion = torch.zeros(self.out_type.size, trivials)
                
                p, c = 0, 0
                for r in self.out_type:
                    pi = 0
                    for irr in r.irreps:
                        irr = self.out_type.fibergroup.irrep(*irr)
                        if irr.is_trivial():
                            bias_expansion[p:p+r.size, c] = torch.tensor(r.change_of_basis[:, pi])
                            c += 1
                        pi += irr.size
                    p += r.size
                
                self.register_buffer("bias_expansion", bias_expansion)
                self.bias = Parameter(torch.zeros(trivials), requires_grad=True)
                self.register_buffer("expanded_bias", torch.zeros(out_type.size))
            else:
                self.bias = None
                self.expanded_bias = None
        else:
            self.bias = None
            self.expanded_bias = None

        # TODO support `groups` arg for conv
        if groups != 1:
            raise NotImplementedError(f'`groups !=1` not supported yet!')

        # BlocksBasisSampler: submodule which takes care of building the filter
        self._basissampler = BlocksBasisSampler(in_type.representations, out_type.representations,
                                                    self._build_kernel_basis,
                                                    basis_filter=basis_filter,
                                                    recompute=recompute)

        if self.basissampler.dimension() == 0:
            raise ValueError('''
                The basis for the steerable filter is empty!
                Tune the `frequencies_cutoff`, `kernel_size`, `rings` or `basis_filter` parameters to allow
                for a larger basis.
            ''')

        self.weights = Parameter(torch.zeros(self.basissampler.dimension()), requires_grad=True)

    @abstractmethod
    def _build_kernel_basis(self, in_repr: Representation, out_repr: Representation) -> KernelBasis:
        raise NotImplementedError

    @property
    def basissampler(self) -> BlocksBasisSampler:
        r"""
        Submodule which takes care of building the filter.

        It uses the learnt ``weights`` to expand a basis and returns a filter in the usual form used by conventional
        convolutional modules.
        It uses the learned ``weights`` to expand the kernel in the G-steerable basis and returns it in the shape
        :math:`(c_\text{out}, c_\text{in}, s^d)`, where :math:`s` is the ``kernel_size`` and :math:`d` is the
        dimensionality of the base space.

        """
        return self._basissampler

    def expand_bias(self) -> torch.Tensor:
        r"""

        Expand the bias in terms of :class:`~escnn.nn._RdPointConv.bias`.

        Returns:
            the expanded bias

        """
        if self.bias is None:
            _bias = None
        else:
            _bias = self.bias_expansion @ self.bias

        return _bias

    def expand_filter(self, points: Union[torch.Tensor, Dict[Tuple, torch.Tensor]]) -> torch.Tensor:
        r"""

        Expand the filter in terms of :class:`~escnn.nn._RdPointConv.weights`.

        Returns:
            the expanded filter sampled on the input points

        """

        return self.basissampler(self.weights, points)

    def expand_parameters(self, points: Union[torch.Tensor, Dict[Tuple, torch.Tensor]]) -> Tuple[torch.Tensor, torch.Tensor]:
        r"""

        Expand the filter in terms of the :attr:`~escnn.nn._RdConv.weights` and the
        expanded bias in terms of :class:`~escnn.nn._RdConv.bias`.

        Returns:
            the expanded filter and bias

        """

        _filter = self.expand_filter(points)
        _bias = self.expand_bias()

        return _filter, _bias

    def forward(self, x: GeometricTensor, edge_index: torch.Tensor, edge_delta: torch.Tensor = None):
        r"""
        Convolve the input with the expanded filter and bias.

        """
        assert isinstance(x, GeometricTensor)
        assert x.type == self.in_type

        assert len(edge_index.shape) == 2
        assert edge_index.shape[0] == 2

        if edge_delta is None:
            pos = x.coords
            row, cols = edge_index
            edge_delta = pos[row] - pos[cols]

        out = self.propagate(edge_index, x=x.tensor, edge_delta=edge_delta)

        if not self.training:
            _bias = self.expanded_bias
        else:
            # retrieve the the bias
            _bias = self.expand_bias()

        if _bias is not None:
            out += _bias

        out = GeometricTensor(out, self.out_type, coords=x.coords)
        
        return out

    def message(self, x_j: torch.Tensor, edge_delta: torch.Tensor=None) -> torch.Tensor:
        return self.basissampler.compute_messages(self.weights, x_j, edge_delta)

    def train(self, mode=True):
        r"""

        If ``mode=True``, the method sets the module in training mode and discards the :attr:`~escnn.nn._RdConv.filter`
        and :attr:`~escnn.nn._RdConv.expanded_bias` attributes.

        If ``mode=False``, it sets the module in evaluation mode. Moreover, the method builds the filter and the bias
        using the current values of the trainable parameters and store them in :attr:`~escnn.nn._RdConv.filter` and
        :attr:`~escnn.nn._RdConv.expanded_bias` such that they are not recomputed at each forward pass.

        .. warning ::

            This behaviour can cause problems when storing the :meth:`~torch.nn.Module.state_dict` of a model while in
            a mode and lately loading it in a model with a different mode, as the attributes of this class change.
            To avoid this issue, we recommend converting the model to eval mode before storing or loading the state
            dictionary.

        Args:
            mode (bool, optional): whether to set training mode (``True``) or evaluation mode (``False``).
                                   Default: ``True``.

        """

        if mode:
            if hasattr(self, "expanded_bias"):
                del self.expanded_bias
        elif self.training:
            # avoid re-computation of the filter and the bias on multiple consecutive calls of `.eval()`

            _bias = self.expand_bias()

            if _bias is not None:
                self.register_buffer("expanded_bias", _bias)
            else:
                self.expanded_bias = None

        return super(_RdPointConv, self).train(mode)

    def evaluate_output_shape(self, input_shape: Tuple) -> Tuple:
        assert len(input_shape) == 3
        assert input_shape[1] == self.in_type.size

        return input_shape[0], self.out_type.size, input_shape[2]

    def __repr__(self):
        extra_lines = []
        extra_repr = self.extra_repr()
        if extra_repr:
            extra_lines = extra_repr.split('\n')

        main_str = self._get_name() + '('
        if len(extra_lines) == 1:
            main_str += extra_lines[0]
        else:
            main_str += '\n  ' + '\n  '.join(extra_lines) + '\n'

        main_str += ')'
        return main_str

    def extra_repr(self):
        s = ('{in_type}, {out_type}, kernel_size={kernel_size}, stride={stride}')
        if self.groups != 1:
            s += ', groups={groups}'
        if self.bias is None:
            s += ', bias=False'
        return s.format(**self.__dict__)

    def check_equivariance(self, atol: float = 1e-5, rtol: float = 1e-6, assertion: bool = True, verbose: bool = True):
    
        # np.set_printoptions(precision=5, threshold=30 *self.in_type.size**2, suppress=False, linewidth=30 *self.in_type.size**2)
    
        P = 30
    
        pos = torch.randn(P, self.d)
        x = torch.randn(P, self.in_type.size)
        x = GeometricTensor(x, self.in_type, pos)
    
        distance = torch.norm(pos.unsqueeze(1) - pos, dim=2, keepdim=False)
    
        thr = sorted(distance.view(-1).tolist())[int(P**2//16)]
        edge_index = torch.nonzero(distance < thr).T.contiguous()

        errors = []
    
        for el in self.space.testing_elements:
        
            out1 = self(x, edge_index).transform(el).tensor.detach().numpy()
            out2 = self(x.transform(el), edge_index).tensor.detach().numpy()
        
            errs = np.abs(out1 - out2)
        
            esum = np.maximum(np.abs(out1), np.abs(out2))
            esum[esum == 0.0] = 1
        
            relerr = errs / esum
        
            if verbose:
                print(el)
                print(relerr.max(), relerr.mean(), relerr.var(), errs.max(), errs.mean(), errs.var())
        
            tol = rtol * esum + atol
        
            if np.any(errs > tol) and verbose:
                print(out1[errs > tol])
                print(out2[errs > tol])
                print(tol[errs > tol])
        
            if assertion:
                assert np.all(
                    errs < tol), 'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}'.format(
                    el, errs.max(), errs.mean(), errs.var())
        
            errors.append((el, errs.mean()))
    
        return errors







from escnn.nn import FieldType, GeometricTensor
from escnn.group import Representation
from escnn.kernels import KernelBasis

from torch_geometric.data import Data

from .rd_point_convolution import _RdPointConv

from typing import Callable, Tuple, Dict, Union, List

import torch
import numpy as np


import math


__all__ = ["R2PointConv"]


class R2PointConv(_RdPointConv):
    
    def __init__(self,
                 in_type: FieldType,
                 out_type: FieldType,
                 groups: int = 1,
                 bias: bool = True,
                 sigma: Union[List[float], float] = None,
                 width: float = None,
                 n_rings: int = None,
                 frequencies_cutoff: Union[float, Callable[[float], int]] = None,
                 rings: List[float] = None,
                 basis_filter: Callable[[dict], bool] = None,
                 recompute: bool = False,
                 ):

        basis_filter, self._rings, self._sigma, self._maximum_frequency = compute_basis_params(
            frequencies_cutoff, rings, sigma, width, n_rings, basis_filter
        )

        super(R2PointConv, self).__init__(
            in_type, out_type,
            d=2,
            groups=groups,
            bias=bias,
            basis_filter=basis_filter,
            recompute=recompute
        )

    def _build_kernel_basis(self, in_repr: Representation, out_repr: Representation) -> KernelBasis:
        return self.space.build_kernel_basis(in_repr, out_repr,
                                             self._sigma, self._rings,
                                             maximum_frequency=self._maximum_frequency
                                             )


def bandlimiting_filter(frequency_cutoff: Union[float, Callable[[float], float]]) -> Callable[[dict], bool]:
    r"""

    Returns a method which takes as input the attributes (as a dictionary) of a basis element and returns a boolean
    value: whether to preserve that element (true) or not (false)

    If the parameter ``frequency_cutoff`` is a scalar value, the maximum frequency allowed at a certain radius is
    proportional to the radius itself. in thi case, the parameter ``frequency_cutoff`` is the factor controlling this
    proportionality relation.

    If the parameter ``frequency_cutoff`` is a callable, it needs to take as input a radius (a scalar value) and return
    the maximum frequency which can be sampled at that radius.

    args:
        frequency_cutoff (float): factor controlling the bandlimiting

    returns:
        a function which checks the attributes of individual basis elements and chooses whether to discard them or not

    """

    if isinstance(frequency_cutoff, float):
        frequency_cutoff = lambda r, fco=frequency_cutoff: r * frequency_cutoff

    def bl_filter(attributes: dict) -> bool:
        return math.fabs(attributes["irrep:frequency"]) <= frequency_cutoff(attributes["radius"])

    return bl_filter


def compute_basis_params(
        frequencies_cutoff: Union[float, Callable[[float], float]] = None,
        rings: List[float] = None,
        sigma: List[float] = None,
        width: float = None,
        n_rings: int = None,
        custom_basis_filter: Callable[[dict], bool] = None,
):

    assert (width is not None and n_rings is not None) != (rings is not None)

    # by default, the number of rings equals half of the filter size
    if rings is None:
        assert width > 0.
        assert n_rings > 0
        rings = torch.linspace(0, width, n_rings)
        rings = rings.tolist()

    if sigma is None:
        sigma = [0.6] * (len(rings) - 1) + [0.4]
        for i, r in enumerate(rings):
            if r == 0.:
                sigma[i] = 0.005
    elif isinstance(sigma, float):
        sigma = [sigma] * len(rings)

    if frequencies_cutoff is None:
        frequencies_cutoff = 3.

    if isinstance(frequencies_cutoff, float):
        frequencies_cutoff = lambda r, fco=frequencies_cutoff: fco * r

    # check if the object is a callable function
    assert callable(frequencies_cutoff)

    maximum_frequency = int(max(frequencies_cutoff(r) for r in rings))

    fco_filter = bandlimiting_filter(frequencies_cutoff)

    if custom_basis_filter is not None:
        basis_filter = lambda d, custom_basis_filter=custom_basis_filter, fco_filter=fco_filter: (
                custom_basis_filter(d) and fco_filter(d)
        )
    else:
        basis_filter = fco_filter

    return basis_filter, rings, sigma, maximum_frequency





import numpy as np

from .basis import KernelBasis
from .steerable_basis import SteerableKernelBasis, IrrepBasis
from .spaces import SpaceIsomorphism
from .wignereckart_solver import WignerEckartBasis

from escnn.group import Representation

from typing import Type, Union, Tuple, Dict, List, Iterable, Callable, Set


class SparseSteerableBasis(KernelBasis):
    
    def __init__(self,
                 X: SpaceIsomorphism,
                 in_repr: Representation,
                 out_repr: Representation,
                 sigma: float,
                 harmonics: List = None,
                 change_of_basis: np.ndarray = None,
                 attributes: Dict = None):
        r"""

        Class which implements a steerable basis for a discrete homogeneous space ```X``` by by embedding the finite
        number of points of ```X``` into :math:`\R^n` and then "diffusing" the kernel basis defined over these
        points in the ambient space using a Gaussian kernel.

        This class only supports finite homogeneous spaces generated by a finite symmetry group, i.e. not only
        :math:`X=G/H` must have a finite number of elements but also :math:`G` itself should.

        Args:
            X:
            in_repr:
            out_repr:
            sigma:
            harmonics:
            change_of_basis:

        """

        self.X = X
        assert in_repr.group == out_repr.group
        self.group = in_repr.group
        self.in_repr = in_repr
        self.out_repr = out_repr

        # check that the homogeneous space has a finite number of elements
        # unfortunately, we can not directly check the size of the homogeneous space G/H so we check that G itself has
        # only a finite number of elements
        assert self.X.G.order() > 0

        points = np.concatenate([
            self.X.projection(g).reshape(-1, 1) for g in self.X.G.elements
        ], axis=1)

        _, idx = np.unique(points.round(decimals=4), axis=1, return_index=True)
        points = points[:, idx]

        assert points.shape == (self.X.dim, self.X.G.order() / self.X.H.order())

        assert sigma > 0., sigma

        self.points = points
        self.sigma = sigma

        self.wigner_basis = SteerableKernelBasis(self.X, in_repr, out_repr, WignerEckartBasis, harmonics=harmonics)

        self._harmonics = self.wigner_basis.compute_harmonics(self.points)

        dim = len(self.wigner_basis)

        super(SparseSteerableBasis, self).__init__(dim, (out_repr.size, in_repr.size))

        if change_of_basis is not None:
            assert change_of_basis.shape == (self.X.dim, self.X.dim)
            # check that the matrix is invertible
            assert np.isfinite(np.linalg.cond(change_of_basis))

            self.points = change_of_basis @ self.points

        self.change_of_basis = change_of_basis

        self._attributes = attributes if attributes is not None else dict()

    def sample(self, points: np.ndarray, out: np.ndarray = None) -> np.ndarray:
        r"""

        Sample the continuous basis elements on the discrete set of points in ``points``.
        Optionally, store the resulting multidimentional array in ``out``.

        ``points`` must be an array of shape `(d, N)`, where `N` is the number of points.

        Args:
            points (~numpy.ndarray): points where to evaluate the basis elements
            out (~numpy.ndarray, optional): pre-existing array to use to store the output

        Returns:
            the sampled basis

        """
    
        assert len(points.shape) == 2
        S = points.shape[1]
        assert points.shape[0] == self.X.dim
    
        if out is None:
            out = np.empty((self.shape[0], self.shape[1], self.dim, S))
    
        assert out.shape == (self.shape[0], self.shape[1], self.dim, S)

        weights = np.expand_dims(self.points, 2) - np.expand_dims(points, 1)
        assert weights.shape == (self.X.dim, self.points.shape[1], S)

        weights = (weights**2).sum(axis=0) / self.sigma**2
        weights = np.exp(- 0.5 * weights)

        B = 0
        harmonics = {}
        outs = {}
        for b, j in enumerate(self.wigner_basis.js):
    
            harmonics[j] = np.einsum('rmi,io->rmo', self._harmonics[j], weights)
            
            outs[j] = out[:, :, B:B + self.dim_harmonic(j), :]
            B += self.dim_harmonic(j)

        self.sample_harmonics(harmonics, outs)
        return out
    
    def sample_harmonics(self, points: Dict[Tuple, np.ndarray], out: Dict[Tuple, np.ndarray] = None) -> Dict[Tuple, np.ndarray]:
        return self.wigner_basis.sample_harmonics(points, out)

    def dim_harmonic(self, j: Tuple) -> int:
        return self.wigner_basis.dim_harmonic(j)
    
    def attrs_j_iter(self, j: Tuple) -> Iterable:
        for attr in self.wigner_basis.attrs_j_iter(j):
            attr['sigma'] = self.sigma
            attr.update(**self._attributes)
            yield attr

    def attrs_j(self, j: Tuple, idx) -> Dict:
        attr = self.wigner_basis.attrs_j(j, idx)

        attr['sigma'] = self.sigma
        attr.update(**self._attributes)
        return attr

    def __getitem__(self, idx):
        attr = self.wigner_basis[idx]

        attr['sigma'] = self.sigma
        attr.update(**self._attributes)
        return attr

    def __iter__(self):
        for attr in self.wigner_basis:
            attr['sigma'] = self.sigma
            attr.update(**self._attributes)
            yield attr

    def __eq__(self, other):
        if not isinstance(other, SparseSteerableBasis):
            return False
        elif self.wigner_basis != other.wigner_basis:
            return False
        elif not np.isclose(self.sigma, other.sigma):
            return False
        elif (self.change_of_basis is None) != (other.change_of_basis is None):
            return False
        elif self.change_of_basis is not None:
            return np.allclose(self.change_of_basis, other.change_of_basis)
        else:
            return True

    def __hash__(self):
        return 100 * hash(self.wigner_basis) * 10 + hash(self.sigma) + hash(str(self.change_of_basis))




import numpy as np

from .steerable_basis import IrrepBasis
from .spaces import SpaceIsomorphism

from escnn.group import *

from typing import Union, Tuple, Dict, Iterable, List
from collections import defaultdict
from itertools import chain

__all__ = [
    "WignerEckartBasis",
    "RestrictedWignerEckartBasis"
]


class WignerEckartBasis(IrrepBasis):
    
    def __init__(self,
                 X: SpaceIsomorphism,
                 in_irrep: Union[str, IrreducibleRepresentation, int],
                 out_irrep: Union[str, IrreducibleRepresentation, int],
                 harmonics: List[Tuple] = None,
                 ):
        r"""
        
        Solves the kernel constraint for a pair of input and output :math:`G`-irreps over an orbit :math:`X` of
        :math:`G` by using the Wigner-Eckart theorem described in
        `A Wigner-Eckart Theorem for Group Equivariant Convolution Kernels <https://arxiv.org/abs/2010.10952>`_ .
        
        Note that the orbit :math:`X` is isomorphic to an homogeneous space :math:`G/H`, for a particular stabilizer
        subgroup :math:`H < G`.
        Hence, the input `X` is an instance of :class:`~escnn.kernels.SpaceIsomorphism` which precisely implements
        such isomorphism.
        
        A basis for functions over :math:`X\cong G/H` is automatically generated via :class:`escnn.group.HomSpace.basis`,
        and can be band-limited by selecting only the subspaces transforming according to the :math:`G`-irreps in the
        input list `harmonics`.
        
        Args:
            X (SpaceIsomorphism): the orbit of :math:`G`
            in_repr (IrreducibleRepresentation): the input irrep
            out_repr (IrreducibleRepresentation): the output irrep
            harmonics (list, optional): the list of :math:`G` irreps to consider for band-limiting
            
        """
        
        group = X.G

        in_irrep = group.irrep(*group.get_irrep_id(in_irrep))
        out_irrep = group.irrep(*group.get_irrep_id(out_irrep))
        
        assert in_irrep.group == X.G
        assert out_irrep.group == X.G
        assert in_irrep.group == out_irrep.group
        
        self.X = X
        
        self.m = in_irrep.id
        self.n = out_irrep.id
        
        if harmonics is not None:
            _js = []
            for j in harmonics:
                if self.X.dimension_basis(j)[1] == 0:
                    # if the G irrep j has multiplicity 0 in L^2(X), there is no harmonic associated with it
                    # so we can skip it
                    continue
                    
                jJl = group._clebsh_gordan_coeff(self.n, self.m, j).shape[-2]
                if jJl > 0 and self.X.dimension_basis(j)[1] > 0:
                    _js.append((j, jJl))
        else:
            _js = group._tensor_product_irreps(self.m, self.n)
    
            _js = [
                (j, jJl)
                for j, jJl in _js
                if self.X.dimension_basis(j)[1] > 0
            ]

        self._coeff = [
            np.einsum(
                'mnsM,kNM->mnksN',
                group._clebsh_gordan_coeff(self.n, self.m, j), group.irrep(*j).endomorphism_basis()
            ) for j, jJl in _js
        ]
        
        for b, (j, jJl) in enumerate(_js):
            coeff = self._coeff[b]
            assert jJl == coeff.shape[-2]

        dim = 0
        self._dim_harmonics = {}
        self._jJl = {}
        for j, jJl in _js:
            self._dim_harmonics[j] = self.X.dimension_basis(j)[1] * jJl * group.irrep(*j).sum_of_squares_constituents
            self._jJl[j] = jJl
            dim += self._dim_harmonics[j]
        
        js = [j for j, _ in _js]
        
        super(WignerEckartBasis, self).__init__(X, in_irrep, out_irrep, js, dim)

    def sample_harmonics(self, points: Dict[Tuple, np.ndarray], out: Dict[Tuple, np.ndarray] = None) -> Dict[Tuple, np.ndarray]:
        if out is None:
            out = {
                j : np.empty((self.shape[0], self.shape[1], self.dim_harmonic(j), points[j].shape[-1]))
                for j in self.js
            }
    
        for j in self.js:
            if j in out:
                assert out[j].shape == (self.shape[0], self.shape[1], self.dim_harmonic(j), points[j].shape[-1])

        for b, j in enumerate(self.js):
            coeff = self._coeff[b]
            
            jJl = coeff.shape[-2]
        
            Ys = points[j]
            np.einsum(
                'Nnksm,miS->NnksiS',
                coeff, Ys,
                out=out[j].reshape((
                    self.out_irrep.size, self.in_irrep.size, self.X.G.irrep(*j).sum_of_squares_constituents, jJl,
                    Ys.shape[1], Ys.shape[2]
                ))
            )
        
        return out

    def dim_harmonic(self, j: Tuple) -> int:
        if j in self._dim_harmonics:
            return self._dim_harmonics[j]
        else:
            return 0

    def attrs_j_iter(self, j: Tuple) -> Iterable:
        
        if self.dim_harmonic(j) == 0:
            return
        
        idx = self._start_index[j]
        
        j_attr = {
            'irrep:'+k: v
            for k, v in self.group.irrep(*j).attributes.items()
        }
        
        dim = self.X.dimension_basis(j)[1]
        for k in range(self.X.G.irrep(*j).sum_of_squares_constituents):
            for s in range(self._jJl[j]):
                for i in range(dim):
                    attr = j_attr.copy()
                    attr["idx"] = idx
                    attr["j"] = j
                    attr["i"] = i
                    attr["s"] = s
                    attr["k"] = k
                    idx += 1
                
                    yield attr

    def attrs_j(self, j: Tuple, idx) -> Dict:
        assert 0 <= idx < self.dim_harmonic(j)
        
        full_idx = self._start_index[j] + idx

        dim = self.X.dimension_basis(j)[1]

        attr = {
            'irrep:'+k: v
            for k, v in self.group.irrep(*j).attributes.items()
        }
        
        attr["idx"] = full_idx
        attr["j"] = j
        attr["i"] = idx % dim
        attr["s"] = (idx // dim) % self._jJl[j]
        attr["k"] = idx // (dim * self._jJl[j])
        
        return attr

    def __getitem__(self, idx):
        assert 0 <= idx < self.dim
    
        i = idx
        for j in self.js:
            dim = self.dim_harmonic(j)
            if i < dim:
                break
            else:
                i -= dim
        
        return self.attrs_j(j, i)
        
    def __iter__(self):
        return chain(self.attrs_j_iter(j) for j in self.js)

    def __eq__(self, other):
        if not isinstance(other, WignerEckartBasis):
            return False
        elif self.X != other.X or self.in_irrep != other.in_irrep or self.out_irrep != other.out_irrep:
            # TODO check isomorphism too!
            return False
        elif len(self.js) != len(other.js):
            return False
        else:
            for b, (j, i) in enumerate(zip(self.js, other.js)):
                if j!=i or not np.allclose(self._coeff[b], other._coeff[b]):
                    return False
            return True

    def __hash__(self):
        return hash(self.X) + hash(self.in_irrep) + hash(self.out_irrep) + hash(tuple(self.js))

    _cached_instances = {}
    
    @classmethod
    def _generator(cls,
                   X: SpaceIsomorphism,
                   psi_in: Union[IrreducibleRepresentation, Tuple],
                   psi_out: Union[IrreducibleRepresentation, Tuple],
                   harmonics: List[Tuple] = None,
                   **kwargs) -> 'IrrepBasis':
    
        assert len(kwargs) == 0
        
        psi_in = X.G.irrep(*X.G.get_irrep_id(psi_in))
        psi_out = X.G.irrep(*X.G.get_irrep_id(psi_out))
        
        if harmonics is not None:
            _harmonics = tuple(sorted(harmonics))
        else:
            _harmonics = None
        key = (X, psi_in.id, psi_out.id, _harmonics)
        
        if key not in cls._cached_instances:
            cls._cached_instances[key] = WignerEckartBasis(X, in_irrep=psi_in, out_irrep=psi_out, harmonics=harmonics)
        return cls._cached_instances[key]


class RestrictedWignerEckartBasis(IrrepBasis):
    
    def __init__(self,
                 X: SpaceIsomorphism,
                 sg_id: Tuple,
                 in_irrep: Union[str, IrreducibleRepresentation, int],
                 out_irrep: Union[str, IrreducibleRepresentation, int],
                 harmonics: Iterable[Tuple],
                 ):
        r"""

        Solves the kernel constraint for a pair of input and output :math:`G`-irreps over an orbit :math:`X` of a larger
        group :math:`G' > G` by using the generalized Wigner-Eckart theorem described in
        `A Program to Build E(N)-Equivariant Steerable CNNs  <https://openreview.net/forum?id=WE4qe9xlnQw>`_
        (Theorem 2.2).

        Note that the orbit :math:`X` is isomorphic to an homogeneous space :math:`G'/H`, for a particular stabilizer
        subgroup :math:`H < G'`.
        Hence, the input `X` is an instance of :class:`~escnn.kernels.SpaceIsomorphism` which precisely implements
        such isomorphism.

        A basis :math:`\mathcal{B} = \{Y_{j'i'}\}` for functions over :math:`X\cong G'/H` is automatically generated via
        :class:`escnn.group.HomSpace.basis`, and can be band-limited by selecting only the subspaces transforming
        according to the :math:`G'`-irreps in the input list `harmonics`.
        
        The equivariance group :math:`G < G'` is identified by the input id ``sg_id``.

        Args:
            X (SpaceIsomorphism): the orbit of :math:`G'`
            sg_id (tuple): id of :math:`G` as a subgroup of :math:`G'`.
            in_repr (IrreducibleRepresentation): the input `G`-irrep
            out_repr (IrreducibleRepresentation): the output `G`-irrep
            harmonics (list, optional): the list of :math:`G'` irreps to consider for band-limiting

        """

        # the larger group G'
        _G = X.G

        # the base space is a homogeneous space for the larger group G'
        self.X = X

        # the smaller equivariance group G
        G, inclusion, restriction = _G.subgroup(sg_id)
        
        self.group = G
        self.sg_id = sg_id

        in_irrep = G.irrep(*G.get_irrep_id(in_irrep))
        out_irrep = G.irrep(*G.get_irrep_id(out_irrep))

        assert in_irrep.group == G
        assert out_irrep.group == G
        assert in_irrep.group == out_irrep.group
        
        self.m = in_irrep.id
        self.n = out_irrep.id

        # irreps of G in the decomposition of the tensor product of in_irrep and out_irrep
        _js_G = [
            j for j, _ in
            G._tensor_product_irreps(self.m, self.n)
        ]
        
        _js = set()
        _js_restriction = defaultdict(list)
        
        assert harmonics is not None, \
            f'A finite list of harmonics to consider is required when using ' \
            f'`RestrictedWignerEckartbasis` to avoid building infinite dimensional bases'
        
        # for each harmonic j' of X considered
        for _j in harmonics:
            if self.X.dimension_basis(_j)[1] == 0:
                # if the G' irrep j' has multiplicity 0 in L^2(X), there is no harmonic associated with it
                # so we can skip it
                continue
                
            # restrict the corresponding G' irrep j' to G
            _j_G = _G.irrep(*_j).restrict(sg_id)
            
            # for each G irrep j in the tensor product decomposition of in_irrep and out_irrep
            for j in _js_G:
                # irrep-decomposition coefficients of j in j'
                id_coeff = []
                p = 0
                # for each G irrep i in the restriction of j' to G
                for i in _j_G.irreps:
                    size = G.irrep(*i).size
                    # if the restricted irrep contains one of the irreps in the tensor product
                    if i == j:
                        id_coeff.append(
                            _j_G.change_of_basis_inv[p:p+size, :]
                        )
                    
                    p += size
                
                # if the G irrep j appears in the restriction of the harmonic j',
                # store its irrep-decomposition coefficients
                if len(id_coeff) > 0:
                    id_coeff = np.stack(id_coeff, axis=-1)
                    _js.add(_j)
                    _js_restriction[_j].append((j, id_coeff))

        _js = sorted(list(_js))
        
        self._coeff = {}
        self._js_restriction = {}
        for _j in _js:
            Y_size = _G.irrep(*_j).size
            coeff = [
                np.einsum(
                    'nmsM,kNM,NYt->nmkstY',
                    G._clebsh_gordan_coeff(self.n, self.m, j), G.irrep(*j).endomorphism_basis(), id_coeff
                ).reshape((out_irrep.size, in_irrep.size, -1, Y_size))
                for j, id_coeff in _js_restriction[_j]
            ]
            
            self._coeff[_j] = np.concatenate(coeff, axis=2)
            self._js_restriction[_j] = [(j, id_coeff.shape[2]) for j, id_coeff in _js_restriction[_j]]

        dim = sum(self.dim_harmonic(_j) for _j in _js)

        super(RestrictedWignerEckartBasis, self).__init__(X, in_irrep, out_irrep, _js, dim)

    def sample_harmonics(self, points: Dict[Tuple, np.ndarray], out: Dict[Tuple, np.ndarray] = None) -> Dict[
        Tuple, np.ndarray]:
        if out is None:
            out = {
                j: np.empty((self.shape[0], self.shape[1], self.dim_harmonic(j), points[j].shape[-1]))
                for j in self.js
            }
    
        for j in self.js:
            if j in out:
                assert out[j].shape == (self.shape[0], self.shape[1], self.dim_harmonic(j), points[j].shape[-1])
    
        for j in self.js:
            coeff = self._coeff[j]
        
            Ys = points[j]
            np.einsum(
                'NnDm,miS->NnDiS',
                coeff, Ys,
                out=out[j].reshape((
                    self.out_irrep.size, self.in_irrep.size, coeff.shape[2], Ys.shape[1], Ys.shape[2]
                ))
            )
    
        return out

    def dim_harmonic(self, j: Tuple) -> int:
        if j in self._coeff:
            return self.X.dimension_basis(j)[1] * self._coeff[j].shape[2]
        else:
            return 0

    def attrs_j_iter(self, j: Tuple) -> Iterable:
        
        if self.dim_harmonic(j) == 0:
            return
        
        idx = self._start_index[j]

        dim = self.X.dimension_basis(j)[1]
        
        j_attr = {
            'irrep:'+k: v
            for k, v in self.X.G.irrep(*j).attributes.items()
        }

        for _j, _jj in self._js_restriction[j]:
            _jJl = self.group._clebsh_gordan_coeff(self.n, self.m, _j).shape[2]
            K = self.group.irrep(*_j).sum_of_squares_constituents
            
            for k in range(K):
                for s in range(_jJl):
                    for t in range(_jj):
                        for i in range(dim):
                            attr = j_attr.copy()
                            attr["idx"] = idx
                            attr["j"] = j
                            attr["_j"] = _j
                            attr["i"] = i
                            attr["t"] = t
                            attr["s"] = s
                            attr["k"] = k

                            assert idx < self.dim

                            idx += 1
                        
                            yield attr

    def attrs_j(self, j: Tuple, idx) -> Dict:
        assert 0 <= idx < self.dim_harmonic(j)
        
        full_idx = self._start_index[j] + idx

        dim = self.X.dimension_basis(j)[1]
    
        for _j, _jj in self._js_restriction[j]:
            _jJl = self.group._clebsh_gordan_coeff(self.n, self.m, _j).shape[2]
            K = self.group.irrep(*_j).sum_of_squares_constituents
            
            d = _jj * _jJl * K * dim
            
            if idx >= d:
                idx -= d
            else:
                break

        attr = {
            'irrep:'+k: v
            for k, v in self.X.G.irrep(*j).attributes.items()
        }
        attr["idx"] = full_idx
        attr["j"] = j
        attr["_j"] = _j
        attr["i"] = idx % dim
        attr["t"] = (idx // dim) % _jj
        attr["s"] = (idx // (dim * _jj)) % _jJl
        attr["k"] = idx // (dim * _jj * _jJl)
        return attr

    def __getitem__(self, idx):
        assert 0 <= idx < self.dim
    
        i = idx
        for j in self.js:
            dim = self.dim_harmonic(j)
            if i < dim:
                break
            else:
                i -= dim
    
        return self.attrs_j(j, i)

    def __iter__(self):
        return chain(self.attrs_j_iter(j) for j in self.js)

    def __eq__(self, other):
        if not isinstance(other, RestrictedWignerEckartBasis):
            return False
        elif self.X != other.X or self.sg_id != other.sg_id or self.in_irrep != other.in_irrep or self.out_irrep != other.out_irrep:
            # TODO check isomorphism too!
            return False
        elif len(self.js) != len(other.js):
            return False
        else:
            for (j, i) in zip(self.js, other.js):
                if j!=i or not np.allclose(self._coeff[j], other._coeff[i]):
                    return False
            return True

    def __hash__(self):
        return hash(self.X) + hash(self.sg_id) + hash(self.in_irrep) + hash(self.out_irrep) + hash(tuple(self.js))

    _cached_instances = {}
    
    @classmethod
    def _generator(cls,
                   X: SpaceIsomorphism,
                   psi_in: Union[IrreducibleRepresentation, Tuple],
                   psi_out: Union[IrreducibleRepresentation, Tuple],
                   harmonics: List[Tuple] = None,
                   **kwargs) -> 'IrrepBasis':
    
        assert len(kwargs) == 1
        assert 'sg_id' in kwargs

        G, _, _ = X.G.subgroup(kwargs['sg_id'])
        psi_in = G.irrep(*G.get_irrep_id(psi_in))
        psi_out = G.irrep(*G.get_irrep_id(psi_out))

        if harmonics is not None:
            _harmonics = tuple(sorted(harmonics))
        else:
            _harmonics = None
            
        key = (X, psi_in.id, psi_out.id, _harmonics, kwargs['sg_id'])

        if key not in cls._cached_instances:
            cls._cached_instances[key] = RestrictedWignerEckartBasis(
                X,
                sg_id=kwargs['sg_id'],
                in_irrep=psi_in,
                out_irrep=psi_out,
                harmonics=harmonics
            )
        
        return cls._cached_instances[key]




import numpy as np
from abc import ABC, abstractmethod
from typing import List, Union, Tuple, Callable, Dict

from .basis import KernelBasis, EmptyBasisException
from .steerable_basis import SteerableKernelBasis
from .spaces import PointRn

__all__ = [
    'GaussianRadialProfile',
    'SphericalShellsBasis'
]

def cart2pol(points):
    # computes the polar coordinates
    
    cumsum = np.sqrt(np.cumsum(points[::-1, :] ** 2, axis=0)[:0:-1, :])
    
    radii = cumsum[0, :]
    
    angles = np.arccos(points[:-1, :] / cumsum)
    
    mask = points[-1, :] < 0
    angles[-1, mask] = 2 * np.pi - angles[-1, mask]
    
    # the angle at the origin is not well defined
    angles[:, radii.reshape(-1) < 1e-9] = np.nan
    
    return radii, angles


def pol2cart(radii, angles):
    assert len(radii.shape) == 2
    assert len(angles.shape) == 2
    assert radii.shape[0] == 1
    assert angles.shape[0] > 0
    assert radii.shape[1] == angles.shape[1]
    
    points = np.empty((angles.shape[0] + 1, angles.shape[1]))
    
    mask = (radii > 1e-9).reshape(-1)
    points[:, ~mask] = 0.
    
    non_origin_count = mask.sum()
    cos = np.empty((angles.shape[0] + 1, non_origin_count))
    sin = np.empty((angles.shape[0] + 1, non_origin_count))
    
    cos[:-1, :] = np.cos(angles[:, mask])
    cos[-1, :] = 1.
    
    sin[1:, :] = np.sin(angles[:, mask])
    sin[0, :] = 1.
    sin = np.cumprod(sin, axis=0)
    
    points[:, mask] = cos * sin * radii[:, mask]
    
    return points


class GaussianRadialProfile(KernelBasis):
    
    def __init__(self, radii: List[float], sigma: Union[List[float], float]):
        r"""

        Basis for kernels defined over a radius in :math:`\R^+_0`.

        Each basis element is defined as a Gaussian function.
        Different basis elements are centered at different radii (``rings``) and can possibly be associated with
        different widths (``sigma``).

        More precisely, the following basis is implemented:

        .. math::

            \mathcal{B} = \left\{ b_i (r) :=  \exp \left( \frac{ \left( r - r_i \right)^2}{2 \sigma_i^2} \right) \right\}_i

        In order to build a complete basis of kernels, you should combine this basis with a basis which defines the
        angular profile through :class:`~escnn.kernels.SphericalShellsBasis`.


        Args:
            radii (list): centers of each basis element. They should be different and spread to cover all
                domain of interest
            sigma (list or float): widths of each element. Can potentially be different.


        """
        
        if isinstance(sigma, float):
            sigma = [sigma] * len(radii)
        
        assert len(radii) == len(sigma)
        assert isinstance(radii, list)
        
        for r in radii:
            assert r >= 0.
        
        for s in sigma:
            assert s > 0.
        
        super(GaussianRadialProfile, self).__init__(len(radii), (1, 1))
        
        self.radii = np.array(radii).reshape(1, 1, -1, 1)
        self.sigma = np.array(sigma).reshape(1, 1, -1, 1)
    
    def sample(self, radii: np.ndarray, out: np.ndarray = None) -> np.ndarray:
        r"""

        Sample the continuous basis elements on the discrete set of radii in ``radii``.
        Optionally, store the resulting multidimentional array in ``out``.

        ``radii`` must be an array of shape `(1, N)`, where `N` is the number of points.

        Args:
            radii (~numpy.ndarray): radii where to evaluate the basis elements
            out (~numpy.ndarray, optional): pre-existing array to use to store the output

        Returns:
            the sampled basis

        """
        assert len(radii.shape) == 2
        assert radii.shape[0] == 1
        
        if out is None:
            out = np.empty((self.shape[0], self.shape[1], self.dim, radii.shape[1]))
        
        assert out.shape == (self.shape[0], self.shape[1], self.dim, radii.shape[1])
        
        radii = radii.reshape(1, 1, 1, -1)
        
        d = (self.radii - radii) ** 2
        
        out = np.exp(-0.5 * d / self.sigma ** 2, out=out)
        
        return out
    
    def __getitem__(self, r):
        assert r < self.dim
        return {"radius": self.radii[0, 0, r, 0], "sigma": self.sigma[0, 0, r, 0], "idx": r}
    
    def __eq__(self, other):
        if isinstance(other, GaussianRadialProfile):
            return np.allclose(self.radii, other.radii) and np.allclose(self.sigma, other.sigma)
        else:
            return False
    
    def __hash__(self):
        return hash(self.radii.tobytes()) + hash(self.sigma.tobytes())


class SphericalShellsBasis(KernelBasis):
    
    def __init__(self,
                 n: int,
                 angular: SteerableKernelBasis,
                 radial: GaussianRadialProfile,
                 filter: Callable[[Dict], bool] = None
                 ):
        r"""

        Build the tensor product basis of a radial profile basis and an spherical profile basis for kernels over the
        Euclidean space :math:`\R^n`.
        
        The kernel space is spanned by an independent basis for each shell.
        The kernel space over shells with positive radius is spanned the basis defined by the `angular` basis
        (an independent copy of each for each cell).
        The kernel over the shells with zero radius (the origin) is spanned by the `origin` basis.
        
        Given the bases :math:`O = \{o_i\}_i` for the origin, :math:`A = \{a_j\}_j` for the spherical shells and
        :math:`D = \{d_r\}_r` for the radial component (indexed by :math:`r \geq 0`, the radius different rings),
        this basis is defined as

        .. math::
            C = \left\{c_{i,j}(\bold{p}) := d_r(||\bold{p}||) a_j(\hat{\bold{p}}) \right\}_{r>0, j} \cup \{d_0(||\bold{p}||) o_i\}_i

        where :math:`(||\bold{p}||, \hat{\bold{p}})` are the polar coordinates of the point
        :math:`\bold{p} \in \R^n`.
        
        Note that the basis on the origin is represented as a simple `np.ndarray` of 3 dimensions, where the last one
        indexes the basis elements as :math:`i` above.
        
        The radial component is parametrized using :class:`~escnn.kernels.GaussianRadialProfile`.
        

        Args:
            n (int): dimension of the Euclidean base space
            angular (SteerableKernelBasis): the angular basis
            radial (GaussianRadialProfile): the basis for the radial profile
            filter (callable, optional): function used to filter out some basis elements. It takes as input a dict
                describing a basis element and should return a boolean value indicating whether to keep (`True`) or
                discard (`False`) the element. By default (`None`), all basis elements are kept.

        Attributes:
            ~.radial (GaussianRadialProfile): the radial basis
            ~.angular (SteerableKernelBasis): the angular basis
            ~.origin (SteerableKernelBasis): the basis for the origin

        """

        self.n = n
        self.radial = radial
        self.angular = angular
        
        # TODO - create singleton classes "spaces"
        # include Spheres S^d, points, lines, etc...
        # SpaceIsom should be a pair (HomSpace, Space)
        # the _is_point() method should belong to the space
        # here we should assert that the SpaceIsom is associated with the sphere S^{n-1}
        
        sphere = angular.X
        try:
            origin = SteerableKernelBasis(
                PointRn(self.n, sphere.G),
                angular.in_repr,
                angular.out_repr,
                angular._irrep_basis,
                # the origin contains only the frequency 0 harmonic
                harmonics=[sphere.zero_harmonic],
                **angular._irrep__basis_kwargs
            )
        except EmptyBasisException:
            origin = None

        if origin is not None:
            assert angular.shape == origin.shape[:2]
            assert angular.in_repr == origin.in_repr
            assert angular.out_repr == origin.out_repr
            assert angular.group == origin.group

            assert len(origin.js) == 1
            assert origin.js[0] == sphere.zero_harmonic
            assert origin.js[0] in angular.js
            assert (
                    origin.dim_harmonic(sphere.zero_harmonic)
                    ==
                    angular.dim_harmonic(sphere.zero_harmonic)
            )
            assert origin.dim_harmonic(sphere.zero_harmonic) == origin.dim
            
        self.origin = origin

        if filter is not None:
            self._filter = np.zeros(len(self.angular) * len(self.radial))
            _idx_map = []
            i = 0
            for attr1 in self.radial:
                for attr2 in self.angular:
                    attr = dict()
                    attr.update(attr1)
                    attr.update(attr2)
                    
                    if filter(attr):
                        self._filter[i] = 1
                        self._idx_map.append(i)
                    i += 1
                    
            dim = self._filter.sum()
            self._idx_map = np.array(_idx_map)
        else:
            self._filter = None
            self._idx_map = None
            dim = len(self.angular) * len(self.radial)
        
        super(SphericalShellsBasis, self).__init__(dim, (radial.shape[0] * angular.shape[0], radial.shape[1] * angular.shape[1]))
    
    def sample(self, points: np.ndarray, out: np.ndarray = None) -> np.ndarray:
        r"""

        Sample the continuous basis elements on a discrete set of ``points`` in the space :math:`\R^n`.
        Optionally, store the resulting multidimensional array in ``out``.

        ``points`` must be an array of shape `(n, N)` containing `N` points in the space.
        Note that the points are specified in cartesian coordinates :math:`(x, y, z, ...)`.

        Args:
            points (~numpy.ndarray): points in the n-dimensional Euclidean space where to evaluate the basis elements
            out (~numpy.ndarray, optional): pre-existing array to use to store the output

        Returns:
            the sampled basis

        """
        assert len(points.shape) == 2
        assert points.shape[0] == self.n
        
        # computes the polar coordinates
        # radii, angles = cart2pol(points)
        
        radii = np.sqrt((points ** 2).sum(axis=0, keepdims=True))
        
        non_origin_mask = (radii > 1e-99).reshape(-1)
        sphere = points[:, non_origin_mask] / radii[:, non_origin_mask]
        origin = points[:, ~non_origin_mask]

        if out is None:
            out = np.empty((self.shape[0], self.shape[1], self.dim, points.shape[1]))
        
        assert out.shape == (self.shape[0], self.shape[1], self.dim, points.shape[1])
        
        # sample the radial basis
        o1 = self.radial.sample(radii)

        # sample the angular basis
        o2 = np.empty((self.shape[0], self.shape[1], self.angular.dim, points.shape[1]))
        o2.fill(np.nan)

        # where r>0, we sample all frequencies
        o2[..., non_origin_mask] = self.angular.sample(sphere)
        
        # only frequency 0 is sampled at the origin. Other frequencies are set to 0
        if self.origin is not None:
            o2[..., :self.origin.dim, ~non_origin_mask] = self.origin.sample(origin)
            o2[..., self.origin.dim:, ~non_origin_mask] = 0.
        else:
            o2[..., ~non_origin_mask] = 0.
            
        assert not np.isnan(o1).any()
        assert not np.isnan(o2[..., non_origin_mask]).any()
        assert not np.isnan(o2[..., ~non_origin_mask]).any()
        assert not np.isnan(o2).any()

        m, n, a, p = o1.shape
        q, r, b, p = o2.shape
       
        if self._filter is None:
            np.einsum("mnap,qrbp->mqnrabp", o1, o2, out=out.reshape((m, q, n, r, a, b, p)))
            return out.reshape((q * m, n * r, self.dim, p))
        else:
            out[:] = np.einsum("mnap,qrb->mqnrabp", o1, o2).reshape((m * q, n * r, a * b, p))[..., self._filter, :]
            return out
    
    def __getitem__(self, idx):
        assert idx < self.dim
        
        if self._idx_map is None:
            _idx = idx
        else:
            _idx = self._idx_map[idx]
        
        idx1, idx2 = divmod(_idx, self.angular.dim)
        attr1 = self.radial[idx1]
        attr2 = self.angular[idx2]
        
        assert attr1["idx"] == idx1
        assert attr2["idx"] == idx2

        attr = dict()
        attr.update(attr1)
        attr.update(attr2)

        attr["idx"] = idx
        attr["idx1"] = attr1["idx"]
        attr["idx2"] = attr2["idx"]

        return attr

    def __iter__(self):
        idx = 0
        i = 0

        # since this methods return iterables of attributes built on the fly, load all attributes first and then
        # iterate on these lists
        radial_attrs = list(self.radial)
        angular_attrs = list(self.angular)

        for idx1, attr1 in enumerate(radial_attrs):
            for idx2, attr2 in enumerate(angular_attrs):
                if self._filter is None or self._filter[i] == 1:
                    
                    assert attr1["idx"] == idx1
                    assert attr2["idx"] == idx2
                    
                    attr = dict()
                    attr.update(attr1)
                    attr.update(attr2)
                    attr["idx1"] = attr1["idx"]
                    attr["idx2"] = attr2["idx"]
                    attr["idx"] = idx

                    yield attr
                    idx += 1
                i += 1
    
    def __eq__(self, other):
        if isinstance(other, SphericalShellsBasis):
            return (
                    self.n == other.n and
                    self.radial == other.radial and
                    self.angular == other.angular and
                    self.origin == other.origin and
                    self._filter == other._filter
            )
        else:
            return False
    
    def __hash__(self):
        return self.n + hash(self.radial) + hash(self.angular) + hash(self.origin) + hash(self._filter)


if __name__ == "__main__":
    
    for _ in range(100):
        for n in range(2, 6):
            x = np.random.randn(n, 4)
            
            radii, angles = cart2pol(x)
            y = pol2cart(radii, angles)
            
            print(x)
            print(radii)
            print(angles)
            print(y)
            assert np.allclose(x, y)



import numpy as np

from escnn.group import *

from typing import Sequence, Tuple

from abc import ABC, abstractmethod

__all__ = [
    'SpaceIsomorphism',
    'CircleO2',
    'CircleSO2',
    'SphereO3',
    'SphereSO3',
    'PointRn',
    #
    'Icosidodecahedron',
    'Dodecahedron',
    'Icosahedron',
]


class SpaceIsomorphism(ABC):
    
    def __init__(self, dim: int, X: HomSpace, action: Representation):
        r"""
            Abstract class that defines an embedding of the homogeneous space :math:`X \cong G/H` into the
            Euclidean space :math:`\R^\text{dim}`.
            
            ``action`` should be a :class:`~escnn.group.Representation` of the group :math:`G` defining its linear
            action on the Euclidean space :math:`\R^\text{dim}`.
            
            The embedding map implemented by this class should be *equivariant* with respect to the natural action of
            :math:`G` on :math:`X` and the action of :math:`G` on :math:`\R^\text{dim}` via the representation
            ``action``.
            
            An instance of this class can be used for example to define steerable kernels over a homogeneous space and,
            then, use them on orbits of the group of interest in :math:`\R^d`; see for example
            :class:`~escnn.kernels.WignerEckartBasis` or :class:`~escnn.kernels.RestrictedWignerEckartBasis`.

        """
        
        assert isinstance(dim, int) and dim > 0
        
        self._dim = dim
        
        # HomSpace: homogeneous space of the group G
        self.X = X
        
        # Group: the group acting on the homogeneous space
        self.G = X.G
        # Group: the stabilizer group of the homogeneous space
        self.H = X.H
        
        assert action.group == self.G
        assert action.size == self.dim
        
        # Representation: the representation of the group ``G`` defining its action on the Euclidean space
        self.action = action
        
        self._H_trivial_id = self.H.trivial_representation.id
        
    @property
    def dim(self):
        r"""
            Integer defining the dimensionality of the Euclidean space where the homogeneous space is embedded.
            
        """
        return self._dim
    
    @property
    def zero_harmonic(self) -> Tuple:
        r"""
            Id of the irrep of ``G`` corresponding to the trivial representation and, therefore, to the harmonic of
            constant functions over the homogeneous space
        """
        return self.G.trivial_representation.id
        
    def dimension_basis(self, j: Tuple):
        r"""
        The dimensionality of the subspace of functions on the homogeneous space which transform according to the irrep
        of ``G`` identified by the id ``j``.
        This corresponds to the number of "frequency"-``j`` harmonics.
        
        The method returns a tuple containing the size of the irrep `j` and the number of subspaces transforming
        according to this irrep.
        The product of these numbers is the number of "frequency"-``j`` harmonics.
        
        .. seealso::
            :class:`~escnn.group.IrreducibleRepresentation` and
            :class:`~escnn.group.HomSpace`
        

        """
        return self.X.dimension_basis(j, self._H_trivial_id)[:-1]

    @abstractmethod
    def _is_point(self, points: np.ndarray) -> np.ndarray:
        pass
    
    @abstractmethod
    def _section(self, points: np.ndarray) -> Sequence[GroupElement]:
        pass

    @property
    @abstractmethod
    def _origin(self) -> np.ndarray:
        r'''
            The point in the Euclidean space where the origin of the homogeneous space is mapped to.
        '''
        pass

    def _projection(self, g: GroupElement) -> np.ndarray:
        # action(g) @ x_0
        # where x_0 is the projection of the identity element `e`
        return self.action(g) @ self._origin

    def is_point(self, points: np.ndarray) -> np.ndarray:
        r'''
            Check if each point in ``points`` belongs to the embedding of the homogeneous space into
            the Euclidean one.
        '''
        assert len(points.shape) == 2
        assert points.shape[0] == self.dim
        
        mask = self._is_point(points)
        
        assert mask.shape == (points.shape[1], )
        
        return mask

    def section(self, points: np.ndarray) -> Sequence[GroupElement]:
        r'''
            This method should implement the inverse of the embedding map composed with a section of the homogeneous
            space.
            In other words, the method should first map a point in the Euclidean space to a point
            :math:`x \in X\cong G/H` in the homogeneous space.
            One such point is a *coset*, i.e. :math:`x = gH` for some :math:`g \in G`.
            The method should, then, pick and return a representative element :math:`g' \in x=gH` of the coset.
        '''
        assert len(points.shape) == 2
        assert points.shape[0] == self.dim
        return self._section(points)

    def projection(self, g: GroupElement) -> np.ndarray:
        r'''
            This method should implement the projection map from :math:`G` to :math:`X\cong G/H` composed
            with the embedding map the homogeneous space in the Euclidean one.
            In other words, the method should first project a group element :math:`g \in G` to the coset
            :math:`x=gH \in X\cong G/H`.
            Then, the method embeds this point :math:`x \in X` in the Euclidean space.
        '''
        assert g.group == self.X.G
        
        point = self._projection(g)

        point = point.squeeze()
        assert len(point.shape) == 1
        assert point.shape[0] == self.dim
        return point.reshape(-1, 1)
        
    def _automatic_basis(self, points: np.ndarray, j) -> np.ndarray:
        r'''
            Automaticlly generates a basis for the homogeneous space by using the
            :meth:`~escnn.kernels.SpaceIsomorphism.section` method and the basis generated in
            :class:`~escnn.group.HomSpace` from the irreps coefficients of :math:`G`.
        '''
        
        assert len(points.shape) == 2
        assert points.shape[0] == self.dim
        
        S = points.shape[1]
        
        basis = []
        for p in range(S):
            g = self.section(points[:, p:p+1])[0]
            basis.append(
                self.X.basis(g, j, self._H_trivial_id)[..., 0]
            )
        
        return np.stack(basis, axis=2)
    
    def basis(self, points: np.ndarray, j) -> np.ndarray:
        r'''
            Sample all harmonic basis functions of frequency ``j`` on the input ``points``.
            
            By default, the basis is automatically generated by using the
            :meth:`~escnn.kernels.SpaceIsomorphism.section` method and the basis generated in
            :class:`~escnn.group.HomSpace` from the irreps coefficients of :math:`G`.
            
            This method is very general but often relatively inefficient with respect to ad hoc implementations for
            specific homogeneous spaces (e.g. for spherical harmonics).
            We recommend overriding this method in subclasses if such implementations are available.
            
        '''
        return self._automatic_basis(points, j)

    def __hash__(self):
        return self.dim + hash(self.X) + hash(self.action)
    
    def __eq__(self, other):
        if not isinstance(other, SpaceIsomorphism):
            return False
        else:
            # TODO: this checks that the two instances are equivalent, but they could have different implementation of
            # the basis... Do we want to still consider these instances the same?
            return self.dim == other.dim and self.X == other.X and self.action == other.action
    
    def _test_section_consistency(self):
        
        assert self._origin.shape == (self.dim, 1)
        
        for h in self.H.testing_elements():
            hx0 = self.action(self.X._inclusion(h)) @ self._origin
            assert np.allclose(self._origin, hx0)
        
        for _ in range(50):
            g = self.G.sample()
            x = self.projection(g)
            
            _g = self.section(x)[0]
            _x = self.projection(_g)
            
            assert np.allclose(x, _x)

    def _test_custom_basis_consistency(self):
        for psi in self.G.irreps():
            j = psi.id
            
            for _ in range(50):
                g = self.G.sample()
            
                x = self.projection(g)
                
                basis_default = self._automatic_basis(x, j)
                basis_custom = self.basis(x, j)
                
                shape = self.X.dimension_basis(j, self._H_trivial_id)[:-1]
                assert basis_custom.shape[:-1] == shape
                assert basis_default.shape[:-1] == shape
                
                # c = (basis_default * basis_custom).sum(axis=0) / (basis_default ** 2).sum(axis=0)
                # # assert np.allclose(c, 1.)
                # print('-----------------------')
                # print(j)
                # print(c[..., 0])
                # print('-----------------------')
                assert np.allclose(basis_default, basis_custom) #, rtol=5e-2, atol=1e-2)

    def _test_equivariance(self):
        for psi in self.G.irreps():
            j = psi.id
        
            for _ in range(10):
                x = self.projection(self.G.sample())
                
                Yx = self.basis(x, j)[..., 0]
                
                # for _ in range(10):
                #     g = self.G.sample()
                for g in self.G.testing_elements():
                    
                    gx = self.action(g) @ x
                    Ygx = self.basis(gx, j)[..., 0]
                    
                    gYx = psi(g) @ Yx
                    
                    assert np.allclose(gYx, Ygx)


class CircleSO2(SpaceIsomorphism):
    
    def __init__(self):
        super(CircleSO2, self).__init__(2, so2_group().homspace(1), so2_group().irrep(1))
    
    @property
    def _origin(self) -> np.ndarray:
        # The projection x_0 of the identity element `e` is the vector [1, 0]^T
        return np.array([1, 0]).reshape(-1, 1)
    
    def _is_point(self, points: np.ndarray) -> np.ndarray:
        # return (points**2).sum(axis=0) > 1e-8
        return np.isclose((points**2).sum(axis=0), 1.)
        
    def _section(self, points: np.ndarray) -> Sequence[GroupElement]:
        assert self._is_point(points).all()
        
        theta = np.arctan2(points[1, ...], points[0, ...])
        return [self.G.element(t) for t in theta]
    
    def _projection(self, g: GroupElement) -> np.ndarray:
        # return g.to('C')
        theta = g.to('radians')
        
        return np.asarray([np.cos(theta), np.sin(theta)])

    def basis(self, points: np.ndarray, j) -> np.ndarray:
        assert len(points.shape) == 2
        assert points.shape[0] == self.dim
        S = points.shape[1]
        
        j = self.G.get_irrep_id(j)[0]
        
        assert isinstance(j, int)

        assert self.is_point(points).all()
        
        if j == 0:
            return np.ones((1, 1, S))
        
        theta = np.arctan2(points[1, ...], points[0, ...])

        cos = np.cos(j * theta)
        sin = np.sin(j * theta)
        
        return np.stack([cos, sin], axis=0).reshape(2, 1, S)


class CircleO2(SpaceIsomorphism):
    
    def __init__(self, axis: float = np.pi / 2.):
        self.axis = axis
        
        o2 = o2_group()
        
        action = o2.irrep(1, 1)
        action = change_basis(action, action(o2.element((0, axis), 'radians')), name=f'StandardAction|axis=[{axis}]')
        
        super(CircleO2, self).__init__(2, o2.homspace((0., 1)), action)
        
        # super(CircleO2, self).__init__(2, o2.homspace((2*axis, 1)), action)

    @property
    def _origin(self) -> np.ndarray:
        # The projection x_0 of the identity element `e` is the vector [1, 0]^T
        # rotated (counter-clockwise) by `self.axis`
        return np.array([np.cos(self.axis), np.sin(self.axis)]).reshape(-1, 1)

    def _is_point(self, points: np.ndarray) -> np.ndarray:
        # return (points**2).sum(axis=0) > 1e-8
        return np.isclose((points**2).sum(axis=0), 1.)

    def _section(self, points: np.ndarray) -> Sequence[GroupElement]:
        assert self._is_point(points).all()

        theta = np.arctan2(points[1, ...], points[0, ...])
        theta -= self.axis
        return [self.G.element((0, t)) for t in theta]
    
    # def _projection(self, g: GroupElement) -> np.ndarray:
    #     f, theta = g.to('radians')
    #     if f:
    #         theta = theta / 2
    #     else:
    #         theta += self.axis
    #     return np.asarray([np.cos(theta), np.sin(theta)])
    
    def basis(self, points: np.ndarray, j) -> np.ndarray:
        assert len(points.shape) == 2
        assert points.shape[0] == self.dim
        S = points.shape[1]

        assert self.is_point(points).all()
        j = self.G.get_irrep_id(j)
        
        if j == (1, 0):
            return np.zeros((1, 0, S))
        
        if j == (0, 0):
            return np.ones((1, 1, S))

        j = j[1]

        theta = np.arctan2(points[1, ...], points[0, ...])
        
        theta -= self.axis
        
        cos = np.cos(j * theta)
        sin = np.sin(j * theta)
        
        return np.stack([cos, sin], axis=0).reshape(2, 1, S)


from lie_learn.representations.SO3.spherical_harmonics import rsh


class SphereSO3(SpaceIsomorphism):
    
    def __init__(self):
        super(SphereSO3, self).__init__(3, so3_group().homspace((False, -1)), so3_group().standard_representation())

    @property
    def _origin(self) -> np.ndarray:
        # The projection x_0 of the identity element `e` is the vector [0, 0, ..., 0, 1]^T
        # this choice of x_0 is convenient when we use SO(3)
        # so SO(2) subgroup is identitied with rotations around the Z axis
        return np.array([0, 0, 1]).reshape(-1, 1)

    def _is_point(self, points: np.ndarray) -> np.ndarray:
        # return (points**2).sum(axis=0) > 1e-8
        return np.isclose((points**2).sum(axis=0), 1.)

    def _section(self, points: np.ndarray) -> Sequence[GroupElement]:
        # radii = np.sqrt((points ** 2).sum(0))
        # assert (radii > 1e-9).all()
        assert self._is_point(points).all()
        radii = 1.

        S = points.shape[1]

        x, y, z = points

        theta = np.arccos(np.clip(z / radii, -1., 1.))
        phi = np.arctan2(y, x)

        # rot = np.asarray([0., theta, phi])
        return [self.G.element((0, theta[s], phi[s]), 'zyz') for s in range(S)]

    def basis(self, points: np.ndarray, l) -> np.ndarray:
        assert len(points.shape) == 2
        assert points.shape[0] == self.dim
        S = points.shape[1]
        
        l = self.G.get_irrep_id(l)[0]
        
        assert isinstance(l, int)

        # radii = np.sqrt((points ** 2).sum(0))
        # assert (radii > 1e-9).all()
        assert self._is_point(points).all()
        radii = 1.

        x, y, z = points
        angles = np.empty((2, S))
        angles[0, :] = np.arccos(np.clip(z / radii, -1., 1.))
        angles[1, :] = np.arctan2(y, x)

        yl = np.empty((2 * l + 1, 1, S))
        for m in range(-l, l + 1):
            yl[m + l, 0, :] = rsh(l, m, np.pi - angles[0, :], angles[1, :])
        
        # the central column of the Wigner D Matrices is proportional to the corresponding Spherical Harmonic
        # we need to correct by this proportion factor
        yl *= np.sqrt(4*np.pi / (2*l+1))
        if l % 2 == 1:
            yl *= -1
        
        return yl


class SphereO3(SpaceIsomorphism):
    
    def __init__(self):
        super(SphereO3, self).__init__(3, o3_group().homspace(('cone', -1)), o3_group().standard_representation())

    @property
    def _origin(self) -> np.ndarray:
        # The projection x_0 of the identity element `e` is the vector [0, 0, ..., 0, 1]^T
        # this choice of x_0 is convenient when we use SO(3)
        # so SO(2) subgroup is identitied with rotations around the Z axis
        return np.array([0, 0, 1]).reshape(-1, 1)

    def _is_point(self, points: np.ndarray) -> np.ndarray:
        # return (points**2).sum(axis=0) > 1e-8
        return np.isclose((points**2).sum(axis=0), 1.)

    def _section(self, points: np.ndarray) -> Sequence[GroupElement]:
        # radii = np.sqrt((points ** 2).sum(0))
        # assert (radii > 1e-9).all()
        assert self._is_point(points).all()
        radii = 1.
        
        S = points.shape[1]
        
        x, y, z = points
        
        theta = np.arccos(np.clip(z / radii, -1., 1.))
        phi = np.arctan2(y, x)
        
        # rot = np.asarray([0., theta, phi])
        return [self.G.element((0, (0, theta[s], phi[s])), 'zyz') for s in range(S)]
    
    def basis(self, points: np.ndarray, l) -> np.ndarray:
        assert len(points.shape) == 2
        assert points.shape[0] == self.dim
        S = points.shape[1]
        
        j, l = self.G.get_irrep_id(l)
        
        assert isinstance(l, int)
        assert isinstance(j, int)

        # radii = np.sqrt((points ** 2).sum(0))
        # assert (radii > 1e-9).all()
        assert self._is_point(points).all()
        radii = 1.

        if l % 2 != j:
            return np.empty((2*l+1, 0, S))
        
        x, y, z = points
        angles = np.empty((2, S))
        angles[0, :] = np.arccos(np.clip(z / radii, -1., 1.))
        angles[1, :] = np.arctan2(y, x)
        
        yl = np.empty((2 * l + 1, 1, S))
        for m in range(-l, l + 1):
            yl[m + l, 0, :] = rsh(l, m, np.pi - angles[0, :], angles[1, :])
    
        # the central column of the Wigner D Matrices is proportional to the corresponding Spherical Harmonic
        # we need to correct by this proportion factor
        yl *= np.sqrt(4 * np.pi / (2 * l + 1))
        if l % 2 == 1:
            yl *= -1

        return yl


class PointRn(SpaceIsomorphism):
    
    def __init__(self, n, G: Group):
        r"""
            A space containing a single point (the origin) in :math:`R^n`.
            It is isomorphic to :math:`G/G` for any compact group :math:`G`.
            
        """
        super(PointRn, self).__init__(n, G.homspace(G.subgroup_self_id), directsum([G.trivial_representation]*n))
        
    @property
    def _origin(self) -> np.ndarray:
        # The projection x_0 of the identity element `e` is the vector [0, 0, ..., 0, 0]^T
        return np.zeros((self.dim, 1))

    def _is_point(self, points: np.ndarray) -> np.ndarray:
        # return (points**2).sum(axis=0) > 1e-8
        return points.shape[0] == self._dim and np.isclose((points ** 2).sum(axis=0), 0.)
    
    def _section(self, points: np.ndarray) -> Sequence[GroupElement]:
        assert self._is_point(points).all()
        
        S = points.shape[1]
        return [self.G.identity for s in range(S)]
    
    def basis(self, points: np.ndarray, l) -> np.ndarray:
        assert len(points.shape) == 2
        assert points.shape[0] == self.dim
        S = points.shape[1]
        
        l = self.G.irrep(*self.G.get_irrep_id(l))
        if l == self.G.trivial_representation:
            yl = np.ones((1, 1, S))
            return yl
        else:
            return np.empty((l.size, 0, S))


# Polyhedrons

class FiniteSpaceWithIcosahedralSymmetry(SpaceIsomorphism, ABC):

    def __init__(self, sg_id: Tuple):

        assert sg_id in [(False, 2), (False, 3), (False, 5)]

        G = ico_group()
        super(FiniteSpaceWithIcosahedralSymmetry, self).__init__(3, G.homspace(sg_id), G.standard_representation)

        # retrieve the rotation axis of one of the elements of the stabilizer group to find the point in R^3 which
        # represents the coset of the identity element of G
        axis = self.X._inclusion(self.H.elements[1]).to('Q')[:3]
        self.__origin = axis.reshape(-1, 1) / np.linalg.norm(axis)

        self._points = np.concatenate([
            self.action(g) @ self._origin for g in self.G.elements
        ], axis=1)

        _, idx = np.unique(self._points.round(decimals=5), axis=1, return_index=True)
        self._points = self._points[:, idx]
        self._sections = [self.G.elements[i] for i in idx]

        assert self._points.shape[1] == self.G.order() / self.H.order(), (self._points.shape, self.G.order(), self.H.order())

    @property
    def _origin(self) -> np.ndarray:
        return self.__origin

    def _is_point(self, points: np.ndarray) -> np.ndarray:
        return np.isclose(np.expand_dims(points, 2), np.expand_dims(self._points, 1)).all(axis=0).any(1)

    def _section(self, points: np.ndarray) -> Sequence[GroupElement]:

        pairs = np.isclose(np.expand_dims(points, 2), np.expand_dims(self._points, 1)).all(axis=0)
        assert pairs.any(1).all()

        assert np.allclose(pairs.astype(np.int).sum(1), 1)

        idxs = pairs.argmax(axis=1)

        sections = [self._sections[i] for i in idxs]
        assert len(sections) == points.shape[1]
        return sections


class Icosidodecahedron(FiniteSpaceWithIcosahedralSymmetry):
    def __init__(self):
        super(Icosidodecahedron, self).__init__((False, 2))


class Icosahedron(FiniteSpaceWithIcosahedralSymmetry):
    def __init__(self):
        super(Icosahedron, self).__init__((False, 3))


class Dodecahedron(FiniteSpaceWithIcosahedralSymmetry):
    def __init__(self):
        super(Dodecahedron, self).__init__((False, 5))
